#include <stdio.h>
#include <stdint.h>
#include <malloc.h>
#include "dwarf.h"

uint32_t get_u_leb128(unsigned char **cp){
	uint32_t r=0;
	uint8_t c,shift=0;
	while(true){
		c=**cp;
		r|=((uint32_t)(c&0x7f))<<shift;
		shift+=7;
		(*cp)++;
		if(!(c&0x80)) return r;
	}
}

int32_t get_s_leb128(unsigned char **cp){
	int32_t r=0;
	uint8_t c,shift=0;
	while(true){
		c=**cp;
		r|=((uint32_t)(c&0x7f))<<shift;
		shift+=7;
		(*cp)++;
		if(!(c&0x80)){
			if ((shift <32) && ((r>>shift-1)&1))
			/* sign extend */
			r |= - (1 << shift);
			return r;
		}
	}
}

uint32_t get16(unsigned char **cp){
	uint32_t r;
	r=((*(*cp+0))<<0)+((*(*cp+1))<<8);
	(*cp)+=2;
	return r;
}

uint32_t get32(unsigned char **cp){
	uint32_t r;
	r=((*(*cp+0))<<0)+((*(*cp+1))<<8)+((*(*cp+2))<<16)+((*(*cp+3))<<24);
	(*cp)+=4;
	return r;
}

void decode_debug_line(unsigned char *src, int size, char *fn){
	unsigned char *cp,*endptr,ch;
	int line_range, line_base,minimum_instruction_length;
	int c,standard_opcode_lengths[256];
	int opcode_base,i,opc,len,xopc;
	int r;
	int fileno=1,line=1,adr=0;
	char file[PATH_MAX],dir[PATH_MAX],*dp;

	cp=src;
	endptr=src+size;
	r=get32(&cp);	/* size */
	r=get16(&cp);	/* Version */
	r=get32(&cp); /* prologue length */
	minimum_instruction_length=*(cp++);
	r=*(cp++);	/* default is stmt */
	line_base=*(char*)(cp++);
	line_range=*(cp++);
	opcode_base=*(cp++);
	for(c=1;c<opcode_base;c++){
		i=get_u_leb128(&cp);
		//printf("%d:%d\n",c,i);
		standard_opcode_lengths[c]=i;
	}
	/* directorys */
	while(*cp){
		dp=dir;
		while(ch=*(cp++)){
			*dp++=ch;
		}
		*dp=0;
		//printf("dir=%s\n",dir);
	}
	cp++;
	/* files */
	while(*cp){
		dp=file;
		while(ch=*(cp++)){
			*dp++=ch;
		}
		*dp=0;
		completeFilename(file,fn,dir);
		//printf("file %d=%s\n",filenNameFuel(),file);
		addFileName(file);
		get_u_leb128(&cp);
		get_u_leb128(&cp);
		get_u_leb128(&cp);
	}
	cp++;

	/* OpCodes */
	do{
		opc=*(cp++);
		if(!opc){	/* extended opcode */
			len=*(cp++);
			xopc=*(cp++);
			len--;
			switch(xopc){

				case DW_LNE_end_sequence:
				//printf("end\n");
				break;

				case DW_LNE_set_address:
				adr=get32(&cp);
				//printf("Set Address: %x\n",adr);
				break;

				case 4:
				//printf("Set Discriminator=%d\n",*(cp++));
				break;

				default:
				printf("unknown extended opcode %x: skip %d bytes\n",xopc,len);
				cp+=len;
				break;
			}
		}
		else if(opc<opcode_base){
			switch(opc){

				case DW_LNS_copy:
				//printf("copy\n");
				break;

				case DW_LNS_advance_pc:
				adr+=get_u_leb128(&cp)*minimum_instruction_length;
				//printf("advance pc: %x\n",adr);
				break;

				case DW_LNS_advance_line:
				line+=get_s_leb128(&cp);
				//printf("advance line: %d\n",line);
				break;

				case DW_LNS_set_file:
				fileno=get_u_leb128(&cp);
				//printf("set file: %x\n",fileno);
				line=1;
				break;

				case DW_LNS_set_column:
				printf("set column: %x\n",get_u_leb128(&cp));
				break;

				case DW_LNS_negate_stmt:
				printf("DW_LNS_negate_stmt\n");
				break;

				case DW_LNS_set_basic_block:
				printf("DW_LNS_set_basic_block\n");
				break;

				case DW_LNS_const_add_pc:
				printf("DW_LNS_const_add_pc\n");
				break;

				case DW_LNS_fixed_advance_pc:
				printf("DW_LNS_fixed_advance_pc: %x\n",get_u_leb128(&cp));
				break;

				case DW_LNS_set_prologue_end:
				printf("DW_LNS_set_prologue_end\n");
				break;

				case DW_LNS_set_epilogue_begin:
				printf("DW_LNS_set_epilogue_begin\n");
				break;

				case DW_LNS_set_isa:
				printf("DW_LNS_set_isa: %x\n",get_u_leb128(&cp));
				break;

				default:
				printf("unknown standard opcode %x\n",opc);

				break;
			}
		}
		if(opc>=opcode_base){
			int line_adv,adr_adv;
			line_adv=(opc-opcode_base)%line_range+line_base;
			adr_adv=(opc-opcode_base)/line_range;
			adr+=adr_adv*minimum_instruction_length;
			line+=line_adv;
			//printf("opc=%3d file=%d line=%d adr=%x\n",opc,fileno,line,adr>>1);
			flash[adr>>1].src_file=fileno-1;
			flash[adr>>1].src_line=line;
			flash[adr>>1].flags|=PM_SOURCE;
		}
	}while(cp<endptr);
	//printf("\n");
}
